# FIAIF is an Intelligent firewall, version: $Revision: 1.92 $
#
# description: Automates a packet filtering firewall with iptables.
#
# Script Author: Anders Fugmann <afu at fugmann dot dhs dot org>
# 
# FIAIF is an Intelligent firewall
# Copyright (C) 2002-2003 Anders Peter Fugmann
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Please drop me a line and give me your thoughts on this software.
# Patches are more than welcome.

###############################################################################
# Clear all chains
# Arguments: 0 for close, 1 for open. 
###############################################################################
function iptables_stop ()
{
    local DEBUG=$1
    local POLICY 
    if (( DEBUG == 0 )); then
	POLICY=DROP
    else
	POLICY=ACCEPT
    fi

    echo -n "Clearing all rules: " 
    IPTABLES -F 
    IPTABLES -F -t nat
    IPTABLES -F -t mangle 

    IPTABLES -X 
    IPTABLES -X -t nat 
    IPTABLES -X -t mangle 

    IPTABLES -Z
    IPTABLES -Z -t nat
    IPTABLES -Z -t mangle
    
    #Always set default policy.
    IPTABLES -P INPUT ${POLICY}   
    IPTABLES -P FORWARD ${POLICY}
    IPTABLES -P OUTPUT ${POLICY}
    
    IPTABLES -t nat -P PREROUTING ACCEPT   
    IPTABLES -t nat -P POSTROUTING ACCEPT
    IPTABLES -t nat -P OUTPUT ACCEPT

    IPTABLES -t mangle -P PREROUTING ACCEPT   
    IPTABLES -t mangle -P OUTPUT ACCEPT
    if (( EXTENDED_MANGLE == 1 )); then
	IPTABLES -t mangle -P INPUT ACCEPT
	IPTABLES -t mangle -P FORWARD ACCEPT
	IPTABLES -t mangle -P POSTROUTING ACCEPT
    fi

    echo "Done."    
}


###############################################################################
# setup_default_chains
###############################################################################
function setup_default_chains ()
{
    debug_out "Setting up default chains."
    local POLICY 
    if (( DEBUG == 0 )); then
	POLICY=DROP
    else
	POLICY=ACCEPT
    fi

    IPTABLES -t mangle -N LOG_DROP
    IPTABLES -t mangle -N LOG_ACCEPT
    IPTABLES -t mangle -N LOG_WATCH

    IPTABLES -N RESERVED_SRC
    IPTABLES -N RESERVED_DST
    IPTABLES -N PRIVATE_SRC
    IPTABLES -N PRIVATE_DST

    IPTABLES -N LOG_MARTIAN
    IPTABLES -N LOG_ACCEPT
    IPTABLES -N LOG_REJECT
    IPTABLES -N LOG_DROP
    IPTABLES -N NOLOG_DROP
    IPTABLES -N LOG_INVALID
    IPTABLES -N LOG_MISS 
    IPTABLES -N LOG_ZONE_MISS 
    IPTABLES -N LOG_LIMIT_ACCEPT
    IPTABLES -N LOG_LIMIT_DROP
    IPTABLES -N LOG_LIMIT_REJECT
    IPTABLES -N LOG_SYN
    IPTABLES -N LOG_SCAN
    IPTABLES -N LOG_TEST
    
    if (( VERBOSE == 1 )); then
	local LIMIT
	if [[ -n "${LOG_LIMIT}" || -n "${LOG_BURST}" ]]; then 
	    LIMIT="-m limit --limit ${LOG_LIMIT} --limit-burst ${LOG_BURST}"
	else
	    LIMIT=""
	fi
	local LOG
	if (( ENABLE_ULOG == 1 )); then
	    let ${LOG_LEVEL:=1}
	    LOG="-j ULOG --ulog-nlgroup ${LOG_LEVEL} --ulog-prefix "
	else
	    let ${LOG_LEVEL:=CRIT}
	    LOG="-j LOG --log-level ${LOG_LEVEL} --log-prefix "
	fi

	IPTABLES -t mangle -A LOG_DROP ${LIMIT} ${LOG} "[${LOG_PREFIX}DROP]:"

	IPTABLES -A LOG_MARTIAN ${LIMIT}            ${LOG} "[${LOG_PREFIX}MARTIAN]:" 
	IPTABLES -A LOG_REJECT ${LIMIT}             ${LOG} "[${LOG_PREFIX}REJECT]:"
	IPTABLES -A LOG_DROP ${LIMIT}               ${LOG} "[${LOG_PREFIX}DROP]:"
	IPTABLES -A LOG_INVALID ${LIMIT}            ${LOG} "[${LOG_PREFIX}INVALID]:"
	IPTABLES -A LOG_MISS ${LIMIT}               ${LOG} "[${LOG_PREFIX}MISS]:"
	IPTABLES -A LOG_ZONE_MISS ${LIMIT}          ${LOG} "[${LOG_PREFIX}UNMATCHED_ZONE]:"
	IPTABLES -A LOG_LIMIT_ACCEPT ${LIMIT}       ${LOG} "[${LOG_PREFIX}LIMIT_ACCEPT]:"
	IPTABLES -A LOG_LIMIT_DROP ${LIMIT}         ${LOG} "[${LOG_PREFIX}LIMIT_DROP]:"
	IPTABLES -A LOG_LIMIT_REJECT ${LIMIT}       ${LOG} "[${LOG_PREFIX}LIMIT_REJECT]:"
	IPTABLES -A LOG_SYN ${LIMIT}                ${LOG} "[${LOG_PREFIX}SYN]:"
	IPTABLES -A LOG_SCAN ${LIMIT}               ${LOG} "[${LOG_PREFIX}SCAN]:" 
	IPTABLES -A LOG_TEST ${LIMIT}               ${LOG} "[${LOG_PREFIX}TEST]:" 
    fi

    IPTABLES -t mangle -A LOG_DROP    -j ${POLICY}
    IPTABLES -t mangle -A LOG_ACCEPT  -j ACCEPT
    IPTABLES -t mangle -A LOG_WATCH ${LIMIT} ${LOG} "[${LOG_PREFIX}WATCH]:"

    IPTABLES -A LOG_MARTIAN           -j ${POLICY}
    IPTABLES -A LOG_DROP              -j ${POLICY}
    IPTABLES -A NOLOG_DROP            -j ${POLICY}
    IPTABLES -A LOG_INVALID           -j ${POLICY}
    IPTABLES -A LOG_MISS              -j ${POLICY}
    IPTABLES -A LOG_ZONE_MISS         -j ${POLICY}
    IPTABLES -A LOG_LIMIT_DROP        -j ${POLICY}
    IPTABLES -A LOG_LIMIT_REJECT      -j ${POLICY}
    IPTABLES -A LOG_SYN               -j ${POLICY}
    IPTABLES -A LOG_SCAN              -j ${POLICY}
    IPTABLES -A LOG_TEST              -j ${POLICY}
    IPTABLES -A LOG_LIMIT_ACCEPT      -j ACCEPT
    IPTABLES -A LOG_ACCEPT            -j ACCEPT

    # All reject targets below here.
    if (( DEBUG == 0 )); then
	POLICY=REJECT
    fi
    IPTABLES -A LOG_REJECT            -j ${POLICY}

    # Make reserved chain
    local RESERVED
    # Either a list or a file.
    debug_out "RESERVED_NETWORKS=${RESERVED_NETWORKS}"
    if [[ -f "${CONF_DIR}/${RESERVED_NETWORKS}" ]]; then
	cat "${CONF_DIR}/${RESERVED_NETWORKS}" | while read RESERVED; do
	    if [[ -z "${RESERVED}" || ${RESERVED} == \#* ]]; then
		continue
	    fi
	    IPTABLES -A RESERVED_SRC -s ${RESERVED} -j LOG_MARTIAN
	    IPTABLES -A RESERVED_DST -d ${RESERVED} -j LOG_MARTIAN
	done
    else
	for RESERVED in ${RESERVED_NETWORKS}; do
	    IPTABLES -A RESERVED_SRC -s ${RESERVED} -j LOG_MARTIAN
	    IPTABLES -A RESERVED_DST -d ${RESERVED} -j LOG_MARTIAN
	done
    fi
    local PRIVATE
    debug_out "PRIVATE_NETWORKS=${PRIVATE_NETWORKS}"
    if [[ -f "${CONF_DIR}/${PRIVATE_NETWORKS}" ]]; then
	cat "${CONF_DIR}/${PRIVATE_NETWORKS}" | while read PRIVATE; do
	    if [[ -z "${PRIVATE}" || ${PRIVATE} == \#* ]]; then
		continue
	    fi
	    IPTABLES -A PRIVATE_SRC -s ${PRIVATE} -j LOG_MARTIAN
	    IPTABLES -A PRIVATE_DST -d ${PRIVATE} -j LOG_MARTIAN
	done
    else
	for PRIVATE in ${PRIVATE_NETWORKS}; do
	    IPTABLES -A PRIVATE_SRC -s ${PRIVATE} -j LOG_MARTIAN
	    IPTABLES -A PRIVATE_DST -d ${PRIVATE} -j LOG_MARTIAN
	done
    fi

    debug_out "Setup packets sanity checks"
    IPTABLES -N SANITY
    add_sanity_check SANITY LOG_SCAN

    local DEVICE
    IPTABLES -N CHECK_IP
    for DEVICE in ${DEVICE_LIST}; do
	IPTABLES -N DEV_${DEVICE}_SRC
	IPTABLES -N DEV_${DEVICE}_DST
	IPTABLES -A CHECK_IP -i ${DEVICE} -j DEV_${DEVICE}_SRC
	IPTABLES -A CHECK_IP -o ${DEVICE} -j DEV_${DEVICE}_DST
    done
}


###############################################################################
# Setup initial redirection to chains.
###############################################################################
function chain_redirect ()
{
    debug_out "Global chains"
    local CHAIN
    local PROTOCOL
    local DEVICE
    local PARAM

    # Setup SEND_CHAINS

    for CHAIN in INPUT OUTPUT FORWARD; do
	IPTABLES -A ${CHAIN} -m state --state INVALID -j LOG_INVALID
	IPTABLES -A ${CHAIN} -m state --state RELATED,ESTABLISHED -j LOG_ACCEPT

	IPTABLES -N ${CHAIN}_NEW
	IPTABLES -A ${CHAIN} -m state --state NEW -j ${CHAIN}_NEW

	# Sanity checking of packets.
	IPTABLES -A ${CHAIN}_NEW -j CHECK_IP
	IPTABLES -A ${CHAIN}_NEW -j SANITY

	for DEVICE in ${DEVICE_LIST}; do
	    IPTABLES -N ${CHAIN}_NEW_${DEVICE}
	    case "${CHAIN}" in
                INPUT)
		    PARAM="-i ${DEVICE}"
		    ;;
		OUTPUT | FORWARD)
		    PARAM="-o ${DEVICE}"
		    ;;
	    esac

	    IPTABLES -A ${CHAIN}_NEW ${PARAM} -j ${CHAIN}_NEW_${DEVICE}
	done	    
    done
    for DEVICE in ${DEVICE_LIST}; do
	IPTABLES -N SEND_NEW_${DEVICE}
	# Insert before all other forward rules
	IPTABLES -I FORWARD_NEW -i ${DEVICE} -j SEND_NEW_${DEVICE}
    done

}

###############################################################################
# set_tos
# Params: TOS protocol ports   
###############################################################################
function set_tos () 
{
    local CHAIN=$1
    local TOS=$2
    local PROTOCOL=$3
    local PORTS=$4
    
    if [[ -n "${PORTS}" ]]; then 
	local PORTS_PARAM=""
	if [[ "${PORTS}" == "any" || "${PORTS}" == "ANY" ]]; then
	    IPTABLES -t mangle -A ${CHAIN} -p ${PROTOCOL} \
		-j TOS --set-tos ${TOS}
	else
	    if [[ "${PROTOCOL}" = "icmp" || "${PROTOCOL}" = "ICMP" ]]; then
		IPTABLES -t mangle -A ${CHAIN} -p ${PROTOCOL} \
		    --icmp-type ${PORTS} -j TOS --set-tos ${TOS}
	    else
		IPTABLES -t mangle -A ${CHAIN} -p ${PROTOCOL} \
		    -m multiport --dports ${PORTS} -j TOS --set-tos ${TOS}
		IPTABLES -t mangle -A ${CHAIN} -p ${PROTOCOL} \
		    -m multiport --sports ${PORTS} -j TOS --set-tos ${TOS}
		
	    fi
	fi
    fi
}

###############################################################################
# Read TOS_FILE, and call set_tos   
###############################################################################
function read_tos()
{
    if [[ -n "${TOS_FILE}" ]]; then
	if [[ ! -f ${CONF_DIR}/${TOS_FILE} ]]; then
	    echo Could not locate TOS settings file: ${CONF_DIR}/${TOS_FILE}.
	else
	    source ${CONF_DIR}/${TOS_FILE}

	    # Create the chain.
	    local CHAIN="SET_TOS"
	    IPTABLES -t mangle -N ${CHAIN}

	    local TOS
            # Setup TOS values:
	    for TOS in ${!TOS_*}; do
		debug_out "${TOS}=${!TOS}"
		set_tos ${CHAIN} ${!TOS}
	    done
	    # Route the the chain.
	    if (( EXTENDED_MANGLE == 1 )); then
		IPTABLES -t mangle -A INPUT -j ${CHAIN}
		IPTABLES -t mangle -A FORWARD -j ${CHAIN}
		IPTABLES -t mangle -A OUTPUT -j ${CHAIN}
	    else
		IPTABLES -t mangle -A PREROUTING -j ${CHAIN}
		IPTABLES -t mangle -A OUTPUT -j ${CHAIN}
	    fi
	fi
    fi
}


###############################################################################
# Setup zone LO. This zone is so specialized, 
# so who would change rules on this.
# I still believe that the loopback interface is weired.
###############################################################################
function configure_zone_lo ()
{
    debug_out "Configuring zone: LO"
    
    #Do not allow forwarding.
    IPTABLES -A FORWARD -o lo -j LOG_DROP
    IPTABLES -A FORWARD -i lo -j LOG_DROP    
    
    # Accept any packets.
    IPTABLES -A INPUT -i lo -j ACCEPT

    # Allow lo to send any packets.
    IPTABLES -A OUTPUT -o lo -j ACCEPT

}

###############################################################################
# State valid:
# RETURN=1, if the state file is valid. 
###############################################################################
function state_valid()
{
    if [[ ! -f ${FIAIF_STATE_FILE} ]]; then 
	return 2
    fi

    if [[ ${FIAIF_STATE_FILE} -ot ${CONF_FILE} || 
          ${FIAIF_STATE_FILE} -ot ${CONF_DIR}/${RESERVED_NETS} || \
          ${FIAIF_STATE_FILE} -ot ${CONF_DIR}/${PRIVATE_NETS}  || \
          ${FIAIF_STATE_FILE} -ot ${CONF_DIR}/${TOS_FILE} || \
          ${FIAIF_STATE_FILE} -ot ${VERSION_FILE} ]]; then
	return 1
    fi

    # Go through all zone files
    local ZONE_FILE
    local ZONE
    for ZONE in ${ZONES}; do
	ZONE_FILE=CONF_${ZONE}
	ZONE_FILE=${CONF_DIR}/${!ZONE_FILE}
	if [[ ${FIAIF_STATE_FILE} -ot ${ZONE_FILE} ]]; then
	    return 1
	fi
    done
    return 0
}

###############################################################################
# Main
###############################################################################
function iptables_setup ()
{
    # Damn bash. We really needed this to be in a seperate function, 
    # but declare only declares locally to functions.
    DEVICE_LIST=""

    local DEVICE
    local STRING1 STRING2
    for ZONE in ${ZONES}; do	
	read_zone ${ZONE}
	if (( $? != 0 ));then
	    STRING1="Warning: Device for ${ZONE} is not up."
	    STRING2="Cannot configure zone."
	    echo "${STRING1}"
	    echo "${STRING2}"
	    debug_out "${STRING1}"
	    debug_out "${STRING2}"
	    declare -ri ${ZONE}_AVAILABLE="0"
	    let DEV_ERRORS++
	    continue
	else
	    declare -r ${ZONE}_DEV="${DEV}"
	    declare -r ${ZONE}_IP="${IP}"
	    declare -ri ${ZONE}_DYNAMIC="${DYNAMIC}"
	    declare -r ${ZONE}_BCAST="${BCAST}"
	    declare -ri ${ZONE}_GLOBAL="${GLOBAL}"
	    declare -r ${ZONE}_NETS="${NET} ${NET_EXTRA}" 
	    declare -r ${ZONE}_IPS="${IP} ${IP_EXTRA}" 
	    declare -ri ${ZONE}_AVAILABLE="1"
	    
	    # Make sure that only one DYNAMIC/GLOBAL zone per interface.
	    local GLOBAL_VAR=${DEV}_GLOBAL_LIST
	    local LOCAL_VAR=${DEV}_LOCAL_LIST
	    if (( GLOBAL == 1 || DYNAMIC == 1 )); then
		if [[ -n "${!GLOBAL_VAR}" || -n "${!LOCAL_VAR}" ]]; then
		    print_err "Error: Zone ${ZONE} declared GLOBAL or DYNAMIC, but a zone has already been defined for ${DEV}."
		    print_err "Conflicting zones are: ${ZONE} ${!GLOBAL_VAR} ${!LOCAL_VAR}"
		    declare ${ZONE}_AVAILABLE="0"
		    let DEV_ERRORS++
		fi
		declare ${GLOBAL_VAR}="${ZONE} ${!GLOBAL_VAR}"
	    else
		if [[ -n "${!GLOBAL_VAR}" ]]; then
		    print_err "Error: A zone declared GLOBAL of DYNAMIC has already been declared for interface ${DEV}."
		    print_err "Conflicting zones are: ${ZONE} ${!GLOBAL_VAR}"
		    declare ${ZONE}_AVAILABLE="0"		
		    let DEV_ERRORS++
		fi
		declare ${LOCAL_VAR}="${ZONE} ${!LOCAL_VAR}"
	    fi
	   
	    
	    # Create a device-based list
	    for DEVICE in ${DEVICE_LIST}; do
		if [[ "${DEVICE}" == "${DEV}" ]]; then 
		    break
		fi
	    done
	    if [[ "${DEVICE}" != "${DEV}" ]]; then
		DEVICE_LIST="${DEVICE_LIST} ${DEV}"
	    fi	
	    
	    # Create zone specific chains.
	    zone_chains ${ZONE}
	    
	fi
    done
    # Setup default global chains.
    setup_default_chains

    # Set TOS values
    read_tos
    
    # Configure the loopback device.
    configure_zone_lo
    
    # Setup chain redirection.
    chain_redirect
    
    echo -n "Configuring zone:"
    for ZONE in ${ZONES}; do
	local ZONE_AVL=${ZONE}_AVAILABLE
	ZONE_AVL=${!ZONE_AVL}
	if (( ZONE_AVL == 1 )); then
	    echo -n " ${ZONE}"
	    configure_zone ${ZONE}
	fi
    done
    echo 

    # Log uncatched packets
    debug_out "Catch unmatched packets"
    IPTABLES -A INPUT -j LOG_ZONE_MISS
    IPTABLES -A OUTPUT -j LOG_ZONE_MISS
    IPTABLES -A FORWARD -j LOG_ZONE_MISS

    debug_out "Log martians" 
    for DEVICE in ${DEVICE_LIST}; do
	IPTABLES -A DEV_${DEVICE}_SRC -j LOG_MARTIAN
	IPTABLES -A DEV_${DEVICE}_DST -j LOG_MARTIAN
    done

    return ${DEV_ERRORS}

}
